#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# IMPORTS
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

from __future__ import annotations;

from src.thirdparty.code import *;
from src.thirdparty.types import *;
from src.thirdparty.plots import *;
from src.thirdparty.maths import *;
from src.thirdparty.render import *;

from src.maths.diagrams import *;
from src.maths.sets import *;

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# EXPORTS
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

__all__ = [
    'FunctionDiagramWidget',
];

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# EXPORTS
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class FunctionProperty(Enum):
    NOTHING = '—';
    INJECTIVE = 'injektiv';
    NOT_INJECTIVE = '¬ injektiv';
    SURJECTIVE = 'surjektiv';
    NOT_SURJECTIVE = '¬ surjektiv';
    BOTH = 'injektiv + surjektiv';
    INJECTIVE_NOT_SURJECTIVE = 'injektiv + ¬ surjektiv';
    NOT_INJECTIVE_SURJECTIVE = '¬ injektiv + surjektiv';
    NEITHER = '¬ injektiv + ¬ surjektiv';

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Class
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class FunctionDiagramWidget():
    state: Optional[widgets.Output];
    N: Optional[int];
    N_max: int;
    fnames: list[str];
    setnames: list[str];
    setsfrom: list[list[Any]];
    card: list[int];

    def __init__(
        self,
        fnames: list[str],
        setnames: list[str],
        setsfrom: dict[str, list[Any]] | list[list[Any]] = [],
        N: Optional[int] = None,
    ):
        self.state = None;
        self.N = N;
        self.N_max = len(fnames);
        assert len(setnames) == len(fnames) + 1, f'The number of sets must be {self.N_max+1}.';
        self.fnames = fnames;
        self.setnames = setnames;
        # fix set contents:
        if isinstance(setsfrom, dict):
            default = randomset_alphabet(mode=Letters.ROMAN, shuffle=True, full=True);
            setsfrom = [ setsfrom.get(name, default) for name in setnames ];
        if len(setsfrom) == 0:
            setsfrom = [
                randomset_alphabet(mode=Letters.SYMBOLS, shuffle=True, full=True),
                randomset_alphabet(mode=Letters.ROMAN, shuffle=True, full=True),
                randomset_alphabet(mode=Letters.HEBREW, shuffle=True, full=True),
                randomset_alphabet(mode=Letters.GREEK, shuffle=True, full=True),
            ];
        n = len(setsfrom);
        self.setsfrom = [ setsfrom[k % n] for k, name in enumerate(setnames) ];
        self.card = [ min(3, len(XX)) for XX in self.setsfrom ];

    def run(self):
        if self.N is None:
            control_nr = widgets.IntSlider(description=f'# Funktionen', min=1, max=self.N_max);
            display(control_nr);
        else:
            control_nr = widgets.IntSlider(value=self.N);
        self.state = widgets.interactive_output(self.handler_wrapper, controls={'N': control_nr});
        display(self.state);

    def handler_plot(self, N: int, show_labels: bool, **kwargs):
        self.card[:(N+1)] = [ kwargs[f'card_{k}'] for k in range(N+1) ];
        properties = [ option_to_property(kwargs[f'property_{k}']) for k in range(N)];
        injective = list(map(is_injective, properties));
        surjective = list(map(is_surjective, properties));
        X = [ XX[:card] for XX, card in zip(self.setsfrom[:(N+1)], self.card[:(N+1)]) ];
        comp = Functions(*[
            Function(
                name     = (f'{self.fnames[k]}', f'{self.setnames[k]}', f'{self.setnames[k+1]}'),
                domain   = X[k],
                codomain = X[k+1],
                fct      = random_function(X[k], X[k+1], injective=injective[k], surjective=surjective[k], force=False)
            )
            for k in range(N)
        ]);
        fig, axs = comp.draw(show_labels=show_labels);
        return;

    def handler_wrapper(self, N: int):
        button_refresh = widgets.ToggleButton(description='Neu laden');
        button_show_labels = widgets.Checkbox(
            description = 'Labels anzeigen?',
            value = True,
            style       = {
                'description_width': 'initial',
            },
            visible     = True
        );
        control_nr = widgets.IntSlider(value=N);
        controls_card = [
            widgets.IntSlider(
                value       = card,
                description = f'|{setname}|',
                min         = 1,
                max         = len(XX),
            )
            for setname, XX, card in zip(self.setnames[:(N+1)], self.setsfrom[:(N+1)], self.card[:(N+1)])
        ];

        # controls for properties of functions
        controls_property = [
            widgets.RadioButtons(
                options     = [ e.value for e in possible_properties(all=True) ],
                value       = FunctionProperty.NOTHING.value,
                layout      = {
                    'description_width': 'initial',
                    'width': 'initial',
                },
                disabled    = False,
            )
            for k in range(N)
        ];
        # NOTE: dynamically change the possible properties based on the cardinalities.
        def update_controls_property(change, k: int):
            self.card[k] = change['new'];
            if k > 0:
                m = self.card[k-1];
                n = self.card[k];
                controls_property[k-1].options = [ e.value for e in possible_properties(m, n) ];
            if k < N:
                m = self.card[k];
                n = self.card[k+1];
                controls_property[k].options = [ e.value for e in possible_properties(m, n) ];

        for k, control_card in enumerate(controls_card):
            control_card.observe(
                partial(update_controls_property, k=k),
                names='value',
            );

        ui = widgets.VBox(
            [ button_refresh ] \
            + [ button_show_labels ] \
            + controls_card \
            + [
                widgets.HBox(
                    [
                        widgets.VBox(
                            [
                                widgets.Text(f'Eigenschaften von {fname}'),
                                control_property
                            ],
                            layout = {
                                'description_width': 'initial',
                                'width': 'initial',
                                'display': 'flex',
                                'align_items': 'stretch',
                                'overflow': 'hidden',
                            }
                        )
                        for fname, control_property in zip(self.fnames[:N], controls_property)
                    ],
                )
            ],
        );
        display(ui);
        display(widgets.interactive_output(
            f = self.handler_plot,
            controls = { 'N': control_nr, 'show_labels': button_show_labels } \
                | { f'card_{k}': controls_card[k] for k in range(N+1) } \
                | { f'property_{k}': controls_property[k] for k in range(N) } \
                | { 'refresh': button_refresh }
        ));
        return;

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# AUXILIARY METHODS
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

def option_to_property(option: str) -> Optional[FunctionProperty]:
    try:
        return FunctionProperty(option);
    except:
        return None;

def possible_properties(m: int = 1, n: int = 1, all: bool = False) -> list[FunctionProperty]:
    if not all:
        if m == 1 and n == 1:
            return [
                e for e in FunctionProperty
                if (is_injective(e) not in [False])
                and (is_surjective(e) not in [False])
            ];
        elif m == 1 and n > 1:
            return [
                e for e in FunctionProperty
                if (is_injective(e) not in [False])
                and (is_surjective(e) not in [True])
            ];
        elif n == 1 and m > 1:
            return [
                e for e in FunctionProperty
                if (is_surjective(e) not in [False])
                and (is_injective(e) not in [True])
            ];
        elif m > n:
            return [
                e for e in FunctionProperty
                if (is_injective(e) not in [True])
            ];
        elif n > m:
            return [
                e for e in FunctionProperty
                if (is_surjective(e) not in [True])
            ];
    return [ e for e in FunctionProperty ];

def is_injective(property: Optional[FunctionProperty]) -> Optional[bool]:
    match property:
        case FunctionProperty.INJECTIVE | FunctionProperty.INJECTIVE_NOT_SURJECTIVE | FunctionProperty.BOTH:
            return True;
        case FunctionProperty.NOT_INJECTIVE | FunctionProperty.NOT_INJECTIVE_SURJECTIVE | FunctionProperty.NEITHER:
            return False;
    return None;

def is_surjective(property: Optional[FunctionProperty]) -> Optional[bool]:
    match property:
        case FunctionProperty.SURJECTIVE | FunctionProperty.NOT_INJECTIVE_SURJECTIVE | FunctionProperty.BOTH:
            return True;
        case FunctionProperty.NOT_SURJECTIVE | FunctionProperty.INJECTIVE_NOT_SURJECTIVE | FunctionProperty.NEITHER:
            return False;
    return None;
